Module analysis

    use working_precision
    use resonances
    use Adsolutions
    
Implicit None

real(wp), parameter, private :: Pi = 3.14159265358979323846_wp !pi

Contains

Subroutine IdentifyParameters(vexp,v0,cModes,minRI,maxRI,iRI,nRI,modeTable,ModeSet,k,p,Airy,fCau)

    real(wp), allocatable :: modeTable(:,:,:,:,:) !table that stores x, q, b for quick access
    real(wp), allocatable :: vexp(:) !A list that contains the wavenumbers of the current frame
    integer, allocatable :: ModeSet(:,:) !the overall best mode set
    real(wp) :: Airy(1:100) !Stores the roots to the Airy function A(-z). Hardcoded in.
    real(wp) :: A(k+2,k+2), d(k+2) !the matrix A and vector d
    real(wp) :: p(k+2) !contains the solved parameters (radius, m0, m1, ...)
    real(wp) :: prev(k+2) !prev solved in previous interval
    real(wp) :: q(cModes), b(cModes) !temporary values of q and b
    real(wp) :: v0,v0t !The midpoint of the spectrum -- wavenumber
    real(wp) :: iRI, minRI, maxRI !search parameters
    real(wp) :: cRI !the current RI
    real(wp) :: delta !fraction error 
    integer :: nRI !number of steps in RI
    integer :: cModes !Number of modes in current frame and the starting position in obsModes
    integer :: k   !Number of terms in the calculated refractive index (m0 + m1*(v-v0) + ... + m2*(v-v0)^k)
    integer :: fCau !Use Cauchy RI formula (0:No, 1:Yes)

    integer :: i,j,w,N,L,T
    
    Call midPoint(v0,v0t,vexp,cModes)
    
    w=0
    j=0
    Do while (w.EQ.0) !loop until parameters no longer change within a tolerance or j gets too big
        Do i = 1, cModes
            cRI=fRI(p,k,vexp(i),v0t,fCau) !inital guess at RI (for Cauchy the m1 term is near zero)
            
            N=ModeSet(i,1)
            L=ModeSet(i,2)
            T=ModeSet(i,3)
            
            !populate list of q and b
            !If (cRI.GT.minRI .AND. cRI.LT.maxRI) Then
            !    Call Interpolation(cRI,minRI,iRI,q(i),b(i),nRI,&
            !           modeTable(:,N,L,T,2),modeTable(:,N,L,T,3))
            !Else
                !if the RI is outside the table range then calculate it.
                !maybe just always do this, the time saved by the table method is not too great
            	q(i)=(MieRes(N,Airy(L),cRI+iRi,T)-MieRes(N,Airy(L),cRI-iRi,T))/(2*iRI)
            	b(i)=MieRes(N,Airy(L),cRI,T)-q(i)*cRI  
            !End if
            
        End do
        
        prev=p
        
        Call fillAd(A,d,q,b,vexp,v0t,cModes,k,fCau)
        Call findpar(A,d,p,k)
        
        Do i = 1, k+2
            delta=Abs(p(i)-prev(i))/Abs(p(i))
            If (delta.GT.1.0E-5_wp) Then
                w=0
                Exit
            End if
            w=1
        End do
        
        j=j+1
        
        If (j.GT.100) Then !arbitrary condition for j getting too big
            w=1
        End if

    End do
	
End Subroutine

Subroutine AssignModes(cModes,minL,maxL,minN,maxN,vexp,v0,modeTable,p,&
                       nRad,iRad,minRad,nRI,iRI,minRI,nDis,iDis,minDis,ModeSet,mlow)
        
    real(wp), allocatable :: modeTable(:,:,:,:,:) !table that stores x, q, b for quick access
    real(wp), allocatable :: vexp(:) !A list that contains the wavenumbers of the current frame
    real(wp), allocatable :: p(:) !in this subroutine initial guesses are determined
    real(wp) :: xexp(cModes) !A list that contains the experimental modes in size parameter
    real(wp) :: m1exp(cModes) !stores m1(vexp(i)-v0) to save calculations
    real(wp) :: modeEr(cModes) !Error for the modes
    integer, allocatable :: minN(:) !The minimum mode number for a given mode order
    integer, allocatable :: maxN(:) !The maximum mode number for a given mode order 
    integer, allocatable :: ModeSet(:,:) !the overall best mode set
    integer :: CurrSet(cModes,3) !the best mode set for the current a, m0, m1
    
    real(wp) :: iRad,minRad,iRI,minRI,iDis,minDis !search parameters
    real(wp) :: bEr,cEr !the best error and the current error
    real(wp) :: delta !error for the current mode
    real(wp) :: v0,v0t !The midpoint of the spectrum -- wavenumber
    real(wp) :: x,q !size parameter and slope of size parameter
    integer :: cModes !Number of modes in current frame and the starting position in obsModes
    integer :: minL,maxL !starting and final mode order
    integer :: nRI,nRad,nDis !number of steps in RI, radius, and dispersion
    
    real(wp) :: xmin,xmax !size parameter limit used when calculating adjustable boundaries
    !integer :: minQN(minL:maxL), maxQN(minL:maxL) !adjustable boundaries based on minN & maxN 
    integer :: minchk !used as a check during a loop

    integer :: mlow !starting integer for mode table in RI (equal to 1 for full search)

    real(wp) :: cRad, m0, m1 !current radius, m0 and m1
    integer :: r,m,d,N,L,T,i
    
    !find the midpoint of the spectrum, user defined or calculated (v0=-1)
    Call midPoint(v0,v0t,vexp,cModes)
    
    bEr=huge(0.0_wp)

    !The search with linear dispersion
    Do r = 1, nRad
        cRad = real(r-1)*iRad + minRad
        Do i = 1, cModes
            xexp(i)=2*Pi*cRad*vexp(i)   
        End do

        Do m = mlow, mlow+nRI-1
            m0 = real(m-mlow)*iRI + minRI !not true anymore??? CHECK Doesn't matter because m0 isn't used real(m-mlow)
            Do d = 1, nDis
                m1 = real(d-1)*iDis + minDis
                Do i = 1, cModes
                    !calculate this term now to save a bit of time later
                    m1exp(i)=m1*(vexp(i)-v0t)   
                End do
                
                !set the errors to large values
                modeEr=huge(0.0_wp)
                
                !rad, m0 and m1 are set, now search the mode set
                Do L = minL, maxL !loop over order
                    !Do N = minQN(L), maxQN(L) !loop over modes
                    Do N = minN(L), maxN(L) !loop over modes
                        Do T = 0,1 !loop over TE and TM
                            x=modeTable(m,N,L,T,1)
                            q=modeTable(m,N,L,T,2)
                            Do i = 1, cModes
                                delta = Abs(xexp(i)-(x+q*m1exp(i)))
                                If (delta.LT.modeEr(i)) Then
                                    modeEr(i)=delta
                                    CurrSet(i,1)=N
                                    CurrSet(i,2)=L
                                    CurrSet(i,3)=T
                                End if
                            End do
                        End do
                    End do
                End do
                
                cEr=0.0_wp
                Do i = 1, cModes
                    cEr = cEr + modeEr(i)**2
                End do
                
                !compare the current mode set to current best mode set
                If (cEr.LT.bEr) Then
                    bEr=cEr
                    Do i = 1, cModes
                        ModeSet(i,1)=CurrSet(i,1)
                        ModeSet(i,2)=CurrSet(i,2)
                        ModeSet(i,3)=CurrSet(i,3)
                    End do                        
                    p(1)=cRad
                    p(2)=m0
                    If (size(p).GT.2) Then !k=0 has no dispersion so p(3) doesn't exist
                        p(3)=m1
                	End if
                End if
                
            End do                
        end do
    End do
  
End subroutine

Subroutine midPoint(v0,v0t,vexp,cModes)
    
    !find the midpoint of the spectrum
    real(wp), allocatable :: vexp(:) !A list that contains the wavenumbers of the current frame
    real(wp) :: v0,v0t !The midpoint of the spectrum -- wavenumber
    integer :: cModes !Number of modes in current frame and the starting position in obsModes
    
    integer :: i
    
    If (v0.LT.0.0_wp) Then
        v0t=0.0_wp
        Do i = 1,cModes
            v0t = vexp(i) + v0t
        End do
        v0t=v0t/cModes
    Else
        v0t=v0
    End if
    
End subroutine 

Subroutine Interpolation(RI,minRI,iRI,q,b,nRI,qtab,btab)
    
    !simple average to find values of q & b from a table
    !instead of calculating them everytime
    
    !This is currently not used in the program
            
    integer :: nRI !number of steps in RI used when building mode table
    
    real(wp) :: qtab(0:nRI+1) !the list of q
    real(wp) :: btab(0:nRI+1) !the list of b
    
    real(wp) :: q,b !the interpolated values of q & b
    
    real(wp) :: RI !the current RI
    real(wp) :: minRI
    real(wp) :: iRI
    real(wp) :: fr !the fraction of the one of the components
    real(wp) :: real_m !real value of m
    
    integer :: m
    
    real_m = (RI-minRI)/iRI + 1.0_wp
    
    fr=Mod(real_m,real(Floor(real_m),wp))
    m=Floor(real_m)
    
    q=qtab(m)*(1.0_wp-fr)+qtab(m+1)*(fr)
    b=btab(m)*(1.0_wp-fr)+btab(m+1)*(fr)
    
End subroutine

Function fRI(p,k,v,v0,fCau)
    
    !calculates the RI

	real(wp) :: fRI
    real(wp) :: p(k+2)
    integer :: k
    real(wp) :: v,v0
    integer :: i
    integer :: fCau !Use Cauchy RI formula (0:No, 1:Yes)
    
    fRI = 0.0_wp

    !the p(1) is the radius, so it's not needed
    do i=2,k+2
        if (fCau.EQ.1) then
		    fRI = fRI + p(i)*(v-v0)**(2*(i-2))  !m0 + m1*(v-v0)^2 + m2*(v-v0)^4 + ...   	
        else
            fRI = fRI + p(i)*(v-v0)**(i-2)  !m0 + m1*(v-v0) + m2*(v-v0)^2 + ...  
    	end if
    end do
    
    Return
    
End function fRI

Subroutine AssignKnown(cModes,sPosn,ModeSet,Known,p,minRad,maxRad,minRI,maxRI)

	!If the modes are already known they can be assigned, crude guesses for the parameters are also made

    integer :: cModes,sPosn !Number of modes in current frame and the starting position in obsModes
    integer, allocatable :: ModeSet(:,:) !the best mode set
    integer, allocatable :: Known(:,:) !If the mode assignment is known it can be used as an input parameter, this stores it.
    real(wp), allocatable :: p(:) !contains the solved parameters (radius, m0, m1, ...)

    real(wp) :: minRad, maxRad
    real(wp) :: minRI, maxRI

	integer :: i,j

	Do i = 1,cModes
		j = i + sPosn - 1
		ModeSet(i,1) = Known(j,1)
		ModeSet(i,2) = Known(j,2)
		ModeSet(i,3) = Known(j,3)
	End do
	
	!Make guesses for the parameters, these can be very crude and it will still work (but they can't be zero)
	p(1)=(minRad+maxRad)/2.0_wp
	p(2)=(minRI+maxRI)/2.0_wp

End subroutine

Subroutine Bounds(minRad,maxRad,minRI,maxRI,minL,maxL,minN,maxN,vexp,Airy)

    real(wp), allocatable :: vexp(:) !A list that contains the wavenumbers of the current frame
    integer, allocatable :: minN(:) !The minimum mode number for a given mode order
    integer, allocatable :: maxN(:) !The maximum mode number for a given mode order    
    real(wp) :: Airy(1:100) !Stores the roots to the Airy function A(-z). Hardcoded in.
    real(wp) :: minRad, maxRad !Range of possible radii
    real(wp) :: minRI, maxRI !Range of possible refractive index
    integer :: minL, maxL !starting and final mode order
    real(wp) :: xmin, xmax !the minimum and maximum size paramter
    
    integer :: j
    
    !the max and min size parameter
    xmax = 2*Pi*maxRad*maxval(vexp)
    xmin = 2*Pi*minRad*minval(vexp)
    
    !Find the min and max mode numbers

    do j = minL, maxL
        minN(j) = xtoN(xmin,minRI,Airy(j)) - 1
        
        do while (MieRes(minN(j),Airy(j),minRI,0).GT.xmin .OR. MieRes(minN(j),Airy(j),minRI,1).GT.xmin)
            minN(j)=minN(j)-1
        end do
        
        maxN(j) = xtoN(xmax,maxRI,Airy(j)) + 1       
    
        do while (MieRes(maxN(j),Airy(j),maxRI,0).LT.xmax .OR. MieRes(maxN(j),Airy(j),maxRI,1).LT.xmax)     
            maxN(j)=maxN(j)+1            
        end do
    end do

End Subroutine Bounds

Function xtoN(x,m,Ar)
    
    integer :: xtoN
    
    real(wp) :: Ar !value of the current Airy function root
    real(wp) :: x !the current size parameter
    real(wp) :: m !current RI

    !N is found using the first two terms from the analytical formula for the resonant modes
    !This is obviously only an approximate guess, but that's not that important for how it's used
    xtoN = NInt(m*x - Ar**3/(6**(1.0_wp/3.0_wp)*(-9*Ar**3*m*x + Sqrt(3.0_wp)*Sqrt(2*Ar**9 + 27*Ar**6*m**2*x**2))** &
           (1.0_wp/3.0_wp)) + (-9*Ar**3*m*x + Sqrt(3.0_wp)*Sqrt(2*Ar**9 + 27*Ar**6*m**2*x**2))**(1.0_wp/3.0_wp)/ &
           6**(2.0_wp/3.0_wp) - 0.5_wp)
    
    Return
    
End function

End Module analysis